/* Soot - a J*va Optimization Framework
 * Copyright (C) 2002 Ondrej Lhotak
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

package soot.jimple.paddle;
import soot.util.*;
import soot.Type;

/** Implementation of points-to set using a bit vector.
 * @author Ondrej Lhotak
 */
public final class BitPointsToSet extends PointsToSetInternal {
    public BitPointsToSet( Type type ) {
        super( type );
        bits = new BitVector( PaddleNumberers.v().contextAllocNodeNumberer().size() );
    }
    /** Returns true if this set contains no run-time objects. */
    public final boolean isEmpty() {
        return empty;
    }

    private final boolean superAddAll( PointsToSetReadOnly other, PointsToSetReadOnly exclude ) {
        boolean ret = super.addAll( other, exclude );
        if( ret ) empty = false;
        return ret;
    }

    private final boolean nativeAddAll( BitPointsToSet other, BitPointsToSet exclude ) {
        BitVector mask = null;
        AbsTypeManager typeManager = PaddleScene.v().tm;
        if( !typeManager.castNeverFails( other.getType(), this.getType() ) ) {
            mask = typeManager.get( this.getType() );
        }
        BitVector obits = other.bits;
        BitVector ebits = ( exclude==null ? null : exclude.bits );
        boolean ret = bits.orAndAndNot( obits, mask, ebits );
        if( ret ) empty = false;
        return ret;
    }

    /** Adds contents of other into this set, returns true if this set 
     * changed. */
    public final boolean addAll( PointsToSetReadOnly other,
            PointsToSetReadOnly exclude ) {
        if( other != null && !(other instanceof BitPointsToSet) )
            return superAddAll( other, exclude );
        if( exclude != null && !(exclude instanceof BitPointsToSet) )
            return superAddAll( other, exclude );
        return nativeAddAll( (BitPointsToSet) other, (BitPointsToSet) exclude );
    }
    /** Calls v's visit method on all nodes in this set. */
    public final boolean forall( P2SetVisitor v ) {
        for( BitSetIterator it = bits.iterator(); it.hasNext(); ) {
            v.visit( (ContextAllocNode) PaddleNumberers.v().contextAllocNodeNumberer().get( it.next() ) );
        }
        return v.getReturnValue();
    }
    /** Adds n to this set, returns true if n was not already in this set. */
    public final boolean add( ContextAllocNode n ) {
        if( PaddleScene.v().tm.castNeverFails( n.getType(), type ) ) {
            return fastAdd( n );
        }
        return false;
    }
    /** Returns true iff the set contains n. */
    public final boolean contains( ContextAllocNode n ) {
        //validate(n);
        return bits.get( n.getNumber() );
    }
    public static P2SetFactory getFactory() {
        return new P2SetFactory() {
            public PointsToSetInternal newSet( Type type ) {
                return new BitPointsToSet( type );
            }
        };
    }

    /* End of public methods. */
    /* End of package methods. */

    private boolean fastAdd( ContextAllocNode n ) {
        //validate(n);
        boolean ret = bits.set( n.getNumber() );
        if( ret ) empty = false;
        return ret;
    }

    private BitVector bits = null;
    private boolean empty = true;

    /*
    public static void validate(ContextAllocNode n) {
        ArrayNumberer an = PaddleNumberers.v().contextAllocNodeNumberer();
        if( n.getNumber() != an.get(n) ) throw new RuntimeException();
        if( an.get(n.getNumber()) != n ) throw new RuntimeException();
    }
    */
}

